﻿# grafikfenster.py
# Auf der Grundlage der wxPython Demo
# Version 2019
# einschliesslich Fuellfarbe und Bildexport
# Grafikfenster einfach skalierbar

import wx
import colorsys
from math import radians
import platform
versionsNummer=platform.python_version()[0]
##if versionsNummer=='2': pass
##else: pass

### ---------------------------------------------------
## Die selbst definierte Exception-Klasse
class VerstossGegenSingletonError(Exception):
    def __init__(self):
        pass
    def __str__(self):
        return "Konstruktor nicht direkt aufrufen!"

USE_BUFFER = True
##ggf ersetzen durch ('wxMSW' in wx.PlatformInfo)
## use buffered drawing on Windows

### ---------------------------------------------------
class Zeichenflaeche(wx.Panel):
    """Die Klasse stellt eine Zeichenflaeche fuer Vektorgrafik bereit.
    Sie wendet das Singleton-Muster an, also kein Konstruktoraufruf,
    sondern Methode GibZeichenflaeche()"""
    __zeichenflaeche=None
    @staticmethod
    def GibZeichenflaeche(parent=None):
        if Zeichenflaeche.__zeichenflaeche == None:
            if parent==None:
                raise Exception('"parent" von Zeichenflaeche fehlt!')
            else:
                Zeichenflaeche.__zeichenflaeche=Zeichenflaeche(parent)
        return Zeichenflaeche.__zeichenflaeche

    def __init__(self, parent):
        """Konstruktor
        !nicht direkt aufrufen!"""
        wx.Panel.__init__(self, parent, -1)
        ### Wesentliche Aenderung: ################################
        if Zeichenflaeche.__zeichenflaeche != None:
            raise VerstossGegenSingletonError()
        Zeichenflaeche.__zeichenflaeche = self
        ###########################################################
        self.Bind(wx.EVT_PAINT, self.OnPaint)
        if USE_BUFFER:
            self.Bind(wx.EVT_SIZE, self.OnSize)
        self.objekte = []
        self.__image=None
        self.fg,self.bg='black','white'

    def Setze_fg_bg(self,fg,bg):
        self.fg,self.bg=fg,bg

    def BildExport(self, dateiname):
        '''exportiert im png- oder jpg-Format'''
        self.InitBuffer()
        typ=dateiname.partition('.')[2].upper()
        if not(typ in ('PNG','JPG')): return 'Dateityp png oder jpg'
        self.__image=self._buffer.ConvertToImage()
        bitmap_type={'PNG':wx.BITMAP_TYPE_PNG,
                     'JPG':wx.BITMAP_TYPE_JPEG}[typ]
        try:
            with open(dateiname, 'wb') as bild_datei:
                self.__image.SaveFile(dateiname,bitmap_type)
        except IOError as ioerr:
            return 'Dateifehler: ' + str(ioerr)            

    def OnSize(self, evt):
        """Ereignismethode, die beim Aendern der Groesse
        des fensters aufgerufen wird."""
        self.InitBuffer()
        evt.Skip()

    def OnPaint(self, evt):
        """Ereignismethode fuer das Zeichnen
        Die erste Anweisung ist im Unterschied zur Demo notwendig,
        um unter Windows beim ersten Zeichnen
        den Puffer richtig zu initialisieren"""
        self.InitBuffer()
        if USE_BUFFER:
            dc = wx.BufferedPaintDC(self, self._buffer)
        else:
            dc = wx.PaintDC(self)
            self.gc = self.MakeGC(dc)
            self.Draw(self.gc)

    def InitBuffer(self):
        """Initialisierungsmethode fuer den Zeichenpuffer"""
        sz = self.GetClientSize()
        sz.width = max(1, sz.width)
        sz.height = max(1, sz.height)
        if versionsNummer=='2': self._buffer = wx.EmptyBitmap(sz.width, sz.height, 32)
        else: self._buffer = wx.Bitmap(sz.width, sz.height, 32)        
        dc = wx.MemoryDC(self._buffer)
        dc.SetBackground(wx.Brush(self.bg))
        dc.Clear()
        self.gc = self.MakeGC(dc)
        self.Draw(self.gc)

    def MakeGC(self, dc):
        """erzeugt den 'Graphics-Context'"""
        try:
            gc = wx.GraphicsContext.Create(dc)
        except NotImplementedError:
            dc.DrawText("This build of wxPython does not support the wx.GraphicsContext "
                        "family of classes.",
                        25, 25)
            return None
        self.gc = gc
        return gc

    def Draw(self, gc):
        """Die eigentliche Zeichenmethode
        wxPython transformiert den Zeichenpfad.
        Wenn man mehrfach zeichnen will, muss man daher
        vor einer Transformation den Zustand sichern und
        anschliessend wiederherstellen."""
        font = wx.SystemSettings.GetFont(wx.SYS_DEFAULT_GUI_FONT)
        font.SetWeight(wx.BOLD)
        gc.SetFont(font, wx.BLUE) # wird ignoriert

        # weissen Hintergrund zeichnen
        back = gc.CreatePath()
        gc.PushState()  # save current translation/scale/other state
        back.AddRectangle(0, 0, self.GetClientSize().width, self.GetClientSize().height)
        gc.SetBrush(wx.Brush(self.bg))
        gc.FillPath(back)          # nur Fuellung
        gc.PopState()              # restore saved state

        for objekt in self.objekte:
            if objekt.GibFigur()==NotImplemented: break
            ## Neumodellierung:
            ## 1. sichtbar wird nicht abgefragt
            ## 2. Texte ermoeglichen
            path = gc.CreatePath()
            gc.PushState()
            path.AddPath(objekt.GibFigur())
            f=wx.TheColourDatabase.FindName(objekt.GibFarbe())
            gc.SetPen(wx.Pen(f))
            try:
                istText=objekt.IstText()
            except AttributeError as e:
                istText=False
            if istText:
                if objekt.GibFont()!=None: gc.SetFont(objekt.GibFont(), f)
                else: gc.SetFont(font, f)
                if versionsNummer=='2':
                    ## angle funktioniert nicht entgegen Doku
                    gc.Translate(objekt.GibXPosition(),
                                 objekt.GibYPosition())
                    gc.Rotate(-radians(objekt.GibOrientierung()))
                    gc.DrawText(objekt.GibText(),0,0)
                else:
                    gc.DrawText(objekt.GibText(),
                                objekt.GibXPosition(),
                                objekt.GibYPosition(),
                                radians(objekt.GibOrientierung()))
            else:
                try:
                    fuellFarbe=objekt.GibFuellFarbe()
                except AttributeError as e:
                    fuellFarbe=None
                if fuellFarbe!=None:
                    ff=wx.TheColourDatabase.FindName(objekt.GibFuellFarbe())
                    gc.SetBrush(wx.Brush(ff))
                    gc.DrawPath(path)
                else:
                    gc.StrokePath(path)
            gc.PopState()

    def Zeichne(self, objekt):
        """Erweiterungen für das Raumplanerprojekt:
        Ruft nacheinander das Zeichnen fuer alle Objekte auf."""
        if objekt in self.objekte:
            self.objekte.remove(objekt)
        self.objekte.append(objekt)
        self.Refresh()

    def Entferne(self, objekt):
        """Entfernt ein Objekt aus der Darstellung."""
        if objekt in self.objekte: self.objekte.remove(objekt)
        self.Refresh()

    def GibGC(self):
        """Gibt den aktuellen Graphics-Context zurueck."""
        return self.gc


### ---------------------------------------------------
from wx.py.shell import ShellFrame
from wx.py.filling import FillingFrame
### ---------------------------------------------------
class GrafikFenster(wx.Frame):
    """
    Die Klasse GrafikFenster erzeugt einen Frame.
    """
    def __init__(self, parent, title,pos=(20, 50), size=(800, 840)):
        wx.Frame.__init__(self, parent, -1, title, pos, size)
        if wx.PlatformInfo[1]=='wxMSW':
            self.SetSize((size[0],size[1]+20))
        self.panel = Zeichenflaeche.GibZeichenflaeche(self)
        self.shellFrame = ShellFrame(parent=self, pos=wx.Point(450,100))
        self.fillingFrame = FillingFrame(parent=self, pos=wx.Point(40,500))

    def ZeigeShellFrame(self, zeigen=True):
        """Zeigt die interaktive Shell an."""
        self.shellFrame.Show(zeigen)

    def ZeigeFillingFrame(self, zeigen=True):
        """Zeigt den Monitor an."""
        self.fillingFrame.Show(zeigen)
