# -*- coding: utf-8 -*-
### <tspGA.py>

# Mutationswahrscheinlichkeit und Selektionswahrscheinlichkeit
# sind positive ganze Zahlen zwischen 0 und 100 ( also % )

import random

### ===== Zufallsgenerator-setzen ===================================
# Setzt den Zufallsgenerator
# Das fuehrt bei erneutem Aufruf zur selben Folge
#random.seed(100)

### ===== Erzeugen derStaedtliste ====================================
# ----- globale Variable
# Das Programm verwendet einige globale Variable:
# - Groesse der Bilschirmdarstellung:
bildGroesse = 800
# - Anzahl der Städte:
anzahlStaedte = 20
# ----- eine Stadt ist ein Zahlenpaar
def erzeugeEineStadt():
    return random.randint(1,bildGroesse-2),random.randint(1,bildGroesse-2)
# ----- gibt eine Liste aller Städte zurück
def erzeugeAlleStaedte(anzahlStaedte=anzahlStaedte):
    ## greift auf die globale Variable zu
    staedte = []
    i = anzahlStaedte
    while i>0:
        stadt = erzeugeEineStadt()
        if not(stadt in staedte):
            staedte.append(stadt)
            i-=1
    return staedte
# Eine globale Variable fuer Testzwecke
staedteListe = erzeugeAlleStaedte()
###Tests:
##print(erzeugeEineStadt())
##print(erzeugeAlleStaedte())
### ===== erzeuge-Individuum ========================================
# Erzeugt ein Individuum der Population.
# Konkret in diesem Fall also eine beliebige Reihenfolge der Städte.

def erzeugeIndividuum(staedteListe):
    individuum = []
    staedte=[]+staedteListe # echte Kopie
    while len(staedte)>0:
        i=random.randint(1,len(staedte))-1
        individuum.append(staedte[i])
        staedte.remove(staedte[i])
    return individuum

###Tests:
##staedte=erzeugeAlleStaedte()
##print(staedte)
##print()
##print(erzeugeIndividuum(staedte))

### ===== bewerte ===================================================
# Die Bewertungsfunktion berechnet die Summe der geometrischen Abstände
# (Luftliniendistanz) der Rundtour.
# ----- abstand
# berechnet den Abstand zweier Städte
import math
def abstand(stadt1,stadt2):
    x1,y1=stadt1
    x2,y2=stadt2
    return math.sqrt((x1-x2)**2+(y1-y2)**2)
    
def bewerte(rundtour):
    summe=0
    for i in range(len(rundtour)-1):
        summe+=abstand(rundtour[i],rundtour[i+1])
    summe+=abstand(rundtour[len(rundtour)-1],rundtour[0])
    return summe

###Test:
##staedte=erzeugeAlleStaedte()
##rundtour=erzeugeIndividuum(staedte)
##print(bewerte(rundtour))

### ===== erzeuge-Population ========================================
# Erzeugt die Grundpopulation für den genetischen Prozess.
# Parameter sind: Anzahl der Individuen und die Städteliste
# Wert: Eine Liste von Individuen
def erzeugePopulation(staedteListe,anzahl=5):
    population = []
    for i in range(anzahl):
        population.append(erzeugeIndividuum(staedteListe))
    return population

###Test:
##staedteListe=erzeugeAlleStaedte()
##print(erzeugePopulation(staedteListe))

### ===== aktionP ===================================================
# Hilfsfunktion (Praedikat), die entscheidet, ob eine Aktion stattfindet
# oder nicht. Die Entscheidung hängt von der übergebenen ganzzahligen
# Wahrscheinlichkeit p ab. 0 < p < 100
def aktionP(p):
    return (random.randint(1,100) <= p)

###Test:
##for i in range(10): print(aktionP(30))

### ===== waehle-Positionen =========================================
# Hilfsfunktion
# Wählt zwei verschiedene Zahlen aus {0 .. n-1} fuer die Position des
# Schnitts in den "Genen" aus und gib sie als Liste  zurück. 
# Die Zahlen sind stets geordnet: [klein gross]
# Unzulässige Parameter für n lösen einen Fehler aus!
def waehlePositionen(n):
    if type(n)!=int: return [0,n-1]
    if n<2: return[0,0]
    z1=random.randint(0,n-1)
    z2=random.randint(0,n-1)
    if (0 in[z1,z2]) and (n-1 in[z1,z2]):
        return waehlePositionen(n) # keine Auswahl, noch einmal versuchen
    if z1<z2: return [z1,z2]
    if z2<z1: return [z2,z1]
    return waehlePositionen(n) # gleich, noch einmal versuchen

###Test:
##print(waehlePositionen(30))

### ===== mutiere ===================================================
# Mutiert ein einzelnes Individuum.
# Dazu wird der Teil der Tour zwischen der ersten und der zweiten
# Position einschliesslich umgekehrt eingebaut.
# Individuum -> Individuum

## Beispielversion:
def mutiere(individuum):
    von,bis = waehlePositionen(len(individuum))
    cut=individuum[von:bis]
    cut.reverse()
    return individuum[:von]+cut+individuum[bis:]

###Test:
##staedte=erzeugeAlleStaedte(10)
##rundtour=erzeugeIndividuum(staedte)
##for i in range(5):
##    print(mutiere(rundtour))

### ===== Mutation ==================================================
# Steuert die Mutation für die gesamte Population.
# Die bei der Mutation neu erzeugten Individuen werden der Population
# hinzugefügt, dadurch ändert sich die Gesamtzahl. Deren Anpassung
# erfolgt bei der Selektion.
# Parameter sind: Population, Mutationswahrscheinlichkeit (pMutatioon)
# Wert: Population zuzüglich der Mutierten
def mutation(population, pMutatioon):
    mutierte = []
    for individuum in population:
        if aktionP(pMutatioon):
            mutierte.append(mutiere(individuum))
        else:
            mutierte.append(individuum)
    return population+mutierte

###Test:
##staedteListe=erzeugeAlleStaedte()
##population = erzeugePopulation(staedteListe)
##print(population)
##print()
##population = mutation(population, 40)

### ===== kreuze ====================================================
# Kreuzt zwei einzelne Individuen.
# Das Kreuzen ist beim TSP schwierig umzusetzen, da keine Staedte weg
# fallen duerfen und umgekehrt.
# Das Problem ist, dass eingekreuzte Abschnitte Staedte enthalten, 
# die schon in der Resttour enthalten sind, andere aber z.T. nicht.
# Individuum x Individuum -> (list Individuum Individuum)

def kreuze(individuum1, individuum2):
    vorn,hinten = waehlePositionen(len(individuum1))
    hinten+=1 # also einschliesslich schneiden
    cut1=individuum1[vorn:hinten]
    cut2=individuum2[vorn:hinten]
    neu1 = []
    neu2 = []
    dif1 = []
    dif2 = []
    for stadt in cut1:
        if not(stadt in cut2):
            dif1.append(stadt)
    for stadt in cut2:
        if not(stadt in cut1):
            dif2.append(stadt)
    
    for stadt in individuum1[:vorn]:
        if not(stadt in cut2):
            neu1.append(stadt)
        else:
            neu1.append(dif1.pop(0))
    neu1 += cut2
    for stadt in individuum1[hinten:]:
        if not(stadt in cut2):
            neu1.append(stadt)
        else:
            neu1.append(dif1.pop(0))
    
    for stadt in individuum2[:vorn]:
        if not(stadt in cut1):
            neu2.append(stadt)
        else:
            neu2.append(dif2.pop(0))
    neu2 += cut1
    for stadt in individuum2[hinten:]:
        if not(stadt in cut1):
            neu2.append(stadt)
        else:
            neu2.append(dif2.pop(0))
    return [neu1, neu2]

###Test:
##staedteListe=erzeugeAlleStaedte(5)
##population = erzeugePopulation(staedteListe,2)
##print(population[0],'\n',population[1])
##p0,p1=kreuze(population[0],population[1])
##print(p0,'\n',p1)

### ===== Kreuzen ===================================================
# Steuert das Kreuzen für die gesamte Population.
# Da beim Kreuzen neue Individuen erzeugt und der Population hinzu-
# gefügt werden, ändert sich die Gesamtzahl. Deren Anpassung erfolgt
# bei der Selektion.
# Die zu kreuzenden Individuen werden nach dem Zufallsprinzip ausge-
# wählt.
# Parameter sind: Population, crossing-over-Wahrscheinlichkeit
# Wert:  Population zuzüglich der Gekreuzten
def kreuzen(population, pCrossingOver):
    gekreuzte = []
    n=len(population)
    for individuum in population:
        if aktionP(pCrossingOver):
            anderes = population[random.randint(0,n-1)]
            while anderes == individuum:
                anderes = population[random.randint(0,n-1)]
            gekreuzte+=kreuze(individuum,anderes)
    return population+gekreuzte

###Test:
##staedteListe=erzeugeAlleStaedte()
##population = erzeugePopulation(staedteListe)
##print(population)
##print()
##population = kreuzen(population, 40)
##for i in range(len(population)):
##    print(population[i])

### ===== besserP ===================================================
# Das Prädikat 'besserP' ist das Vergleichsprädikat zur Bewertung und
# ist daher vom speziellen Problem abhängig. Hier wird die vorgegebene
# Funktion bewerte (s.o.) genommen.
def besserP(individuum1, individuum2):
   return bewerte (individuum1) < bewerte (individuum2)

###Test:
##staedteListe=erzeugeAlleStaedte()
##population = erzeugePopulation(staedteListe)
##print(besserP(population[0],population[1]))
##print(besserP(population[1],population[0]))

### ===== selektiere ================================================
# Selektiert von zwei Individuen das mit der höheren Bewertungszahl.
def selektiere(individuum1, individuum2):
    if besserP(individuum1, individuum2):
        return individuum1
    return individuum2

###Test:
##staedteListe=erzeugeAlleStaedte()
##population = erzeugePopulation(staedteListe)
##print(selektiere(population[0],population[1]))
##print(selektiere(population[1],population[0]))

### ===== Selektion =================================================
# Es gibt viele veschiedene Selektionsmechanismen. Beispielsweise
# kann die Auswahl proportional zur Bewertungsfunktion erfolgen.
# Hier werden die vorhandenen Individuen einfach zufällig ausgewählt
# und einer paarweisen Auswahl unterzogen: Selektiert wird aus der
# Menge der Individuen durch direkten Vergleich von je zwei
# Individuen. Der Parameter Gesamtzahl gewährleistet, dass wieder
# die ursprüngliche Populationsgrösse hergestellt wird.
# Dazu wird mit der Selektion abgebrochen, wenn eine genügend grosse
# Zahl (-> Gesamtzahl) ausgewählt wurde.
def selektion(population, gesamtzahl):
    akku = []
    while gesamtzahl > len(akku):
        pos1,pos2 = waehlePositionen(len(population))
        akku.append(selektiere(population[pos1],population[pos2]))
    return akku

#Test:
##staedteListe=erzeugeAlleStaedte()
##population = erzeugePopulation(staedteListe,5)
##print(population)
##print()
##print(selektion(population, 5))

### ===== Bewertung =================================================
# Führt die Bewertung der gesamten Population durch. Dazu wird die 
# Populationsliste in eine Liste mit Paaren [Individuum Bewertung]
# verwandelt.
def bewertung(population):
    akku = []
    for i in range(len(population)):
        akku.append([population[i],bewerte(population[i])])
    return akku

###Test:
##staedteListe=erzeugeAlleStaedte()
##population = erzeugePopulation(staedteListe,5)
##print(bewertung(population))

### ===== bestes-Individuum =========================================
# Für die grafische Darstellung, aber auch für das Ergebnis des GA
# ist es sinnvoll, das 'beste' Individuum bestimmen zu können.
def bestesIndividuum(population):
    indexBestes = 0
    for i in range(1,len(population)):
        if besserP(population[i],population[indexBestes]):
            indexBestes = i
    return population[indexBestes]

###Test:
##staedteListe=erzeugeAlleStaedte()
##population = erzeugePopulation(staedteListe,5)
##for i in range(len(population)):
##    print(population[i]," --> ",bewerte(population[i]))
##print()
##print(bestesIndividuum(population))

### ===== schrittDaten ==============================================
# sichert die Daten der jeweils besten Touren
schrittDaten=[]
def gibSchrittDaten():
    return schrittDaten
def sichereBeste(tour):
    wert=bewerte(tour)
    schrittDaten.append([tour,wert])
def loescheBeste():
    schrittDaten=[]

### ===== Ausgabe ===============================================
# Die Ausgabe der 'besten' Lösungen in der Shell
def ausgabe(individuum):
    print(individuum," --> ",bewerte(individuum))

### ===== genetischer-Algorithmus ===================================
def genetischerAlgorithmus(populationsGroesse,
                           anzahlSchritte,
                           mutationsWahrscheinlichkeit,
                           crossingOverWahrscheinlichkeit):
    population = erzeugePopulation(staedteListe,populationsGroesse)
    ##print(population)
    ##ausgabe(bestesIndividuum(population))
    for i in range(anzahlSchritte):
        population= selektion(kreuzen(mutation(population,
                                               mutationsWahrscheinlichkeit),
                                      crossingOverWahrscheinlichkeit),
                              populationsGroesse)
        sichereBeste(bestesIndividuum(population))
    beste=bestesIndividuum(population)
    return [beste,bewerte(beste)]

###Test:
##staedteListe=erzeugeAlleStaedte(3)
##population = genetischerAlgorithmus(2,10,30,40)
##for i in range(len(population)):
##    print(population[i]," --> ",bewerte(population[i]))
##print()
##print(bestesIndividuum(population))

### ===== GA-Aufruf =================================================
# Eine Aufrufhülle für 'genetischerAlgorithmus', bei der die
# Parameter zur Laufzeit eingegeben werden können
def GA_Aufruf():
    populationsGroesse=int(input("Gib die Zahl der Individuen an [> 0] : "))
    anzahlSchritte=int(input("Wieviel Schritte ?             [> 0] : "))
    mutationsWahrscheinlichkeit=int(input("Mutationswahrscheinlichkeit ?  [00 .. 99] : "))
    crossingOverWahrscheinlichkeit=int(input("Wahrscheinlichkeit für crossing-over ? [00 .. 99] : "))
    genetischerAlgorithmus(populationsGroesse,
                           anzahlSchritte,
                           mutationsWahrscheinlichkeit,
                           crossingOverWahrscheinlichkeit)

### ===== Aufruf ====================================================
##print(genetischerAlgorithmus(20, 20, 40, 40))
